"""Class that encapsulate the docker client and provide common methods"""
import atexit
import functools
import logging
import os
import re
import sys
import warnings
from functools import partial
from multiprocessing import Process
from typing import Any, Callable, Dict, List, Mapping, Optional, Pattern, Tuple, Union

from python_on_whales import Container, Volume, docker
from python_on_whales.exceptions import DockerException

from .config import ODOO_7, ODOO_8, ODOO_9, ODOO_10, Configuration

_logger = logging.getLogger(__name__)

# TODO detect that user is member of docker group

Mount = List[str]


def mount(
    target: str, source: str, mount_type: Optional[str] = None, readonly: bool = False
) -> Mount:
    result = [f"target={target}", f"source={source}"]
    if mount_type is not None:
        result.append(f"type={mount_type}")
    if readonly:
        result.append("readonly")
    return result


class DockerClient:
    """Provide additional abstraction and extra methods of the docker client API"""

    @classmethod
    def create_volume(
        cls, volume_name: str, extra_labels: Optional[Dict[str, str]] = None
    ) -> Tuple[Volume, bool]:
        """Return the volume passed in parameter, creating it if it does not exist.
        :param volume_name: name of the volume to create
        :param extra_labels: extra labels to put on the volume (only on
        creation)
        :return: a tuple, the first one is the volume, the second is a boolean
        indicating that the volume has just been created (true if created)
        """
        volumes = docker.volume.list(filters={"name": volume_name})
        if volumes:
            _logger.debug("Volume %s already exists", volume_name)
            return volumes[0], False
        _logger.debug("Creating volume %s", volume_name)
        labels = dict(odoo_scripts="")
        if extra_labels:
            labels.update(extra_labels)
        return docker.volume.create(volume_name, labels=labels), True

    @classmethod
    def exec(
        cls,
        container: Container,
        command: List[str],
        user: str,
        interactive: bool = False,
    ) -> int:
        """Execute a command in a container

        :param container: Container to exec on
        :param command: Command to be executed
        :param user: User to execute command as.
        :param interactive: True if the container needs to be interactive
        """
        with capture_docker_exit_code() as exit_code:
            container.execute(command, user=user, interactive=interactive, tty=True)
        return int(exit_code)

    @staticmethod
    def _replace_then_print(
        replaces: List[Tuple[Pattern[str], str]], log_line: str, file=sys.stdout
    ):
        # apply all the replacement of the path inside the image to the real path
        for pattern, value in replaces:
            log_line = pattern.sub(value, log_line)
        # print to play nicely with pycharm that will linkify the result
        print(log_line, file=file)

    @classmethod
    def launch(
        cls,
        repository: str,
        tag: str,
        run_kwargs: Dict[str, Any],
        pull: bool = True,
        display_log: bool = True,
    ) -> Container:
        container, stop_remove_registered = cls._create(
            repository, tag, run_kwargs, pull
        )
        _logger.debug("Starting container %s", container.name)
        container.start()
        if display_log:
            process = Process(target=DockerClient.print_log, args=(container,))
            process.start()

            def stop_remove_wait_process():
                stop_remove(container)
                if process and process.is_alive():
                    _logger.info("Waiting for log process to finish")
                    process.join()

            atexit.register(stop_remove_wait_process)
            atexit.unregister(stop_remove_registered)

        return container

    @classmethod
    def run(
        cls,
        repository: str,
        tag: str,
        run_kwargs: Dict[str, Any],
        pull: bool = True,
        target_source_dict: Optional[Mapping[str, str]] = None,
    ) -> int:
        container, stop_remove_registered = cls._create(
            repository, tag, run_kwargs, pull
        )
        _logger.debug("Starting container %s", container.name)
        container.start()

        atexit.unregister(stop_remove_registered)
        remove_registered = functools.partial(remove, container)
        atexit.register(remove_registered)

        # TODO semble trop tard
        replaces: List[Tuple[Pattern[str], str]] = (
            [(re.compile(key), value) for key, value in target_source_dict.items()]
            if target_source_dict
            else []
        )
        cls.print_log(container, replaces)

        _logger.debug("Waiting for container %s", container.name)
        exit_code = docker.wait(container)
        _logger.debug("Removing container %s", container.name)
        container.remove(volumes=True)
        atexit.unregister(remove_registered)
        return exit_code

    @classmethod
    def _create(
        cls,
        repository: str,
        tag: str,
        run_kwargs: Dict[str, Any],
        pull: bool = True,
    ) -> Tuple[Container, Callable]:
        """Run a container with the provided args"""

        image = f"{repository}:{tag}"

        if pull:
            _logger.debug("Pulling %s", image)
            # TODO find a better way to avoid error when pulling without a connection
            try:
                docker.image.pull(image)
            except DockerException:
                _logger.warning(
                    "Issue when pulling, using existing version, if any", exc_info=True
                )

        container: Union[None, Container] = None

        # XXX When pressing ctrl-c during the container creating, the container
        #  is not removed as there is no return of the value of the Container
        #  object. There is several ways to do that:
        #  - give the container a unique name we know, and use that to clean
        #  - maybe use some kind of time frame to identify the container
        #  - warn the user of a potential container leak by using signals

        stop_remove_registered = functools.partial(stop_remove, container)
        atexit.register(stop_remove_registered)

        # split create and start so in case of SIGTERM, container has fewer chances
        # of being None
        _logger.debug("Create container with image %s", image)
        container = docker.container.create(image, **run_kwargs)
        _logger.debug("Created container %s", container.name)
        return container, stop_remove_registered

    @classmethod
    def print_log(
        cls,
        container: Container,
        replaces: Optional[List[Tuple[Pattern[str], str]]] = None,
    ):
        """Print the logs of a container, eventually replacing some of its output."""
        if replaces is None:
            replaces = []
        logs = docker.container.logs(container, follow=True, stream=True)
        if logs is not None and not isinstance(logs, str):
            for channel, log in logs:
                # \r is needed when using a process that also outputs the logs
                log_str = log.decode().rstrip() + "\r"
                cls._replace_then_print(replaces, log_str, getattr(sys, channel))


def get_odoo_base_path(config: Configuration) -> str:
    """Return value of env ODOO_BASE_PATH in the local image name."""
    image = docker.image.list(config.local_image_name)
    if not image:
        raise Exception(
            "Image %s is not built, needed to be inspected", config.local_image_name
        )
    if isinstance(image[0].config.env, list):
        for env in image[0].config.env:
            if env.startswith("ODOO_BASE_PATH="):
                return env[15:]
    raise Exception("Missing ODOO_BASE_PATH in the image")


def modules_mount(config: Configuration) -> Tuple[List[Mount], Dict[str, str]]:
    """Return a list of Mount for the modules in the configuration"""
    modules = config.modules
    _logger.debug("addon modules: %s", ",".join(config.modules))
    project_path = config.path
    mount_list: List[Mount] = []
    mount_dict: Dict[str, str] = {}
    if config.odoo_type in (ODOO_7, ODOO_8, ODOO_9, ODOO_10):
        base_python_path = "/usr/local/lib/python2.7/dist-packages/"
        target_path = "/opt/odoo/sources/odoo/addons"
    else:
        base_python_path = os.path.dirname(get_odoo_base_path(config))
        target_path = get_odoo_base_path(config) + "/addons"

    for module in modules:
        full_target_path = os.path.join(target_path, os.path.basename(module))
        full_project_path = os.path.realpath(os.path.join(project_path, module))
        mount_list.append(mount(full_target_path, full_project_path, "bind"))
        mount_dict[full_target_path] = full_project_path

    _logger.debug("python packages: %s", ",".join(config.python_packages))
    for python_package in config.python_packages.items():
        for dir_ in python_package[1]["dirs"]:
            full_target_path = os.path.join(base_python_path, dir_[1])
            full_project_path = os.path.realpath(
                os.path.join(project_path, python_package[0], dir_[0])
            )
            mount_list.append(mount(full_target_path, full_project_path, "bind"))
            mount_dict[full_target_path] = full_project_path

    return mount_list, mount_dict


def anthem_mounts(config: Configuration) -> List[Mount]:
    """Return a list of mounts for all the anthem songs present in the super project."""
    project_path = os.path.realpath(config.path)
    mounts: List[Mount] = []
    # No anthem on old ODOO
    if config.odoo_type in (ODOO_7, ODOO_8, ODOO_9, ODOO_10):
        return mounts
    # properly defined, so nothing to do
    if "." in config.python_packages:
        return mounts
    # Mount songs for anthem to avoid having to rebuild the image when they are changed.
    # Assume they are in songs/something and put in the image in the same place Odoo is
    # installed.
    if os.path.exists("songs"):
        warnings.warn(
            "Automatic guessing of songs directories will be disabled in the next"
            "version. Adjust your configuration to include them.",
            stacklevel=2,
        )
        for root, dirs, _files in os.walk("songs"):
            for directory in dirs:
                mounts.append(
                    mount(
                        os.path.dirname(get_odoo_base_path(config))
                        + f"/{root}/{directory}",
                        os.path.join(project_path, root, directory),
                        "bind",
                        True,
                    )
                )
    _logger.debug("mounts len: %s", len(mounts))
    return mounts


def odoo_source_mounts(config: Configuration, source_path: str) -> List[Mount]:
    """Return a list of mounts for a directory containing Odoo sources."""
    mounts: List[Mount] = []
    full_source_path = os.path.realpath(source_path)
    if config.odoo_type in (ODOO_7, ODOO_8):
        mounts.append(mount("/opt/odoo/sources/odoo", source_path, "bind"))
    else:
        install_path = (
            "/usr/local/lib/python2.7/dist-packages/odoo"
            if config.odoo_type in (ODOO_9, ODOO_10)
            else get_odoo_base_path(config)
        )
        for entry in os.listdir(os.path.join(full_source_path, "odoo")):
            if entry != "addons":
                mounts.append(
                    mount(
                        install_path + "/" + entry,
                        os.path.join(full_source_path, "odoo", entry),
                        "bind",
                    )
                )
        for entry in os.listdir(os.path.join(full_source_path, "odoo", "addons")):
            # for directory in dirs:
            mounts.append(
                mount(
                    install_path + "/addons/" + entry,
                    os.path.join(full_source_path, "odoo", "addons", entry),
                    "bind",
                )
            )
        # iterate on all the modules, avoids
        for entry in os.listdir(os.path.join(full_source_path, "addons")):
            mounts.append(
                mount(
                    install_path + "/addons/" + entry,
                    os.path.join(full_source_path, "addons", entry),
                    "bind",
                )
            )
    return mounts


class DockerService:
    """Represent a service in a docker and allows common methods to manipulate it."""

    def __init__(
        self,
        name: str,
        repository: str,
        tag: str,
        stop_remove_if_exists: bool,
        immediate_pull: bool = False,
        attach: bool = True,
        stream: bool = True,
        stop_at_exit: bool = True,
        force_pull: bool = False,
    ):
        """
        :param name: name for the container
        :param repository: repository to use for the image
        :param tag: repository tag to use for the image
        :param immediate_pull: pull now (rather than do it later, like during start)
        :param stream: if not attached, indicate to stream output or not
        """
        super().__init__()
        self.name = name
        self.repository: str = repository
        self.tag = tag
        self.base_image = f"{self.repository}:{self.tag}"
        if immediate_pull:
            self.pull_image()
        self.attach = attach
        self.stream = stream
        self.stop_at_exit = stop_at_exit
        # Retrieve the container if it already exists and is running
        self._container: Optional[Container] = None
        if docker.container.exists(self.name):
            self._container = docker.container.list(
                all=True, filters={"name": self.name}
            )[0]
        self._log_process: Optional[Process] = None
        self.stop_remove_if_exists = stop_remove_if_exists
        self.force_pull: bool = force_pull

    def start(self) -> bool:
        """Start the container"""
        # TODO avoid pull if already done with immediate pull
        self.pull_image()
        if self._container is not None:
            # TODO should it crash?
            _logger.info("Container %s already exist", self.name)
            if self.stop_remove_if_exists:
                # stop and remove
                self.stop_and_remove()
            elif not self.is_running:
                _logger.info("Container %s is not running", self.name)
                self.remove()
            else:
                return True
        self._start()
        if self.stop_at_exit:
            atexit.register(partial(self.stop_and_remove, False))
        if self.attach and self._container is not None:
            DockerClient.print_log(self._container)
        elif self.stream:
            self._start_log_process()
        return True

    def _start_log_process(self):
        """Start a process that outputs log from the started container."""
        if not self._log_process:
            process = Process(target=DockerClient.print_log, args=(self._container,))
            process.start()
            self._log_process = process

    def _create_param(self) -> dict:
        return dict(
            name=self.name,
            network_mode="host",
            image=self.base_image,
        )

    def _start(self) -> None:
        self._container = docker.container.create(**self._create_param())
        _logger.info("Starting container %s", self.name)
        self._container.start()

    def stop_and_remove(self, expect_running: bool = True) -> None:
        """Stop and remove the container"""
        try:
            self.stop(expect_running)  # call stop func
            self.remove()  # call remove func
        except DockerException as exception:
            _logger.info("Stop and remove container, got: %s", exception)
            raise exception

    def stop(self, expect_running: bool = True) -> None:
        """Stop service container."""
        try:
            if not self.is_running:
                if expect_running:
                    _logger.warning("Container %s is not running", self.name)
            elif self._container is not None:
                _logger.info("Stopping %s container", self.name)
                self._container.stop()
                docker.container.wait(self._container)
        except DockerException as exception:
            _logger.error("Stopping container %s, got: %s", self.name, exception)
            raise
        self.join_log_process()

    def stop_log_process(self):
        if self._log_process and self._log_process.is_alive():
            self._log_process.terminate()

    def join_log_process(self):
        if self._log_process and self._log_process.is_alive():
            _logger.info("Waiting for log process to finish")
            self._log_process.join()

    def remove(self):
        """Remove the service container created"""
        try:
            if self._container is not None:
                _logger.info("Removing %s container", self.name)
                self._container.remove()
            else:
                _logger.info("Container %s already removed", self.name)
        except docker.errors.NotFound as exception:
            _logger.error("Removing container %s, got: %s", self.name, exception)
            raise exception

    def pull_image(self):
        """Pulls image if needed or forced."""
        if self.force_pull or not docker.image.list(self.base_image):
            _logger.info("Pulling %s", self.base_image)
            docker.image.pull(f"{self.repository}:{self.tag}")

    @property
    def is_running(self) -> bool:
        """Verify the container status
        :return: true if the status is running
        """
        if self._container is None:
            return False
        return bool(self._container.state.running)

    @property
    def is_exited(self) -> bool:
        """
        Verify the rednerd container status
        :return: Boolean if the status is exited
        """
        if self._container is None:
            return True
        return self._container.state.status in [None, "exited"]


def get_volumes(base_dir, directory: str, write: bool = False):
    """return volumes to bind"""
    path = os.path.abspath(os.path.join(base_dir, directory))
    volumes = [
        (
            base_dir,
            base_dir,
            "rw" if write else "ro",
        )
    ]
    # also bind any symbolic link, but only for the given directory or its parent.
    # all symbolic link could be scanned for but that might take too much time
    for potential_link in (path, os.path.dirname(path)):
        if os.path.islink(potential_link):
            volumes.append(
                (
                    potential_link,
                    os.path.realpath(potential_link),
                    "rw" if write else "ro",
                )
            )
    return volumes, path


def remove(container: Union[None, Container]):
    """Remove the container"""
    if container:
        _logger.debug("Removing container %s", container.name)
        container.remove()


def stop_remove(container: Union[None, Container]):
    """Stop then remove the container."""
    if container:
        _logger.debug("Stopping container %s", container.name)
        container.stop()
        _logger.debug("Waiting for container %s", container.name)
        docker.wait(container)
    remove(container)


class _MutableInt:
    """Class with an int.
    Used in context manager to provide exit status."""

    def __init__(self):
        self.value = 0

    def set(self, value: int):
        """Update the value"""
        self.value = value

    def __int__(self) -> int:
        return self.value

    def __bool__(self):
        return bool(self.value)


class capture_docker_exit_code:
    """Context manager to get error code when using docker::run"""

    def __init__(self, print_exc: bool = True):
        self.exit_code = _MutableInt()
        self.print_exc = print_exc

    def __enter__(self):
        return self.exit_code

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type == DockerException:
            self.exit_code = exc_val.return_code
            if self.print_exc:
                if exc_val.stdout:
                    print(exc_val.stdout)
                if exc_val.stderr:
                    print(exc_val.stderr, file=sys.stderr)
            # Need to return a true value to ignore the error
            return True
        return False
