#!/usr/bin/env python3
"""Run pylint in a docker
"""
import argparse
import logging
import sys
from typing import Dict, List, Optional

from .config import Config, Configuration
from .docker_build import add_build_options, build_local_image, get_build_options
from .docker_client import DockerClient, Mount, anthem_mounts, modules_mount, mount
from .parsing import apply, basic_parser

__version__ = "1.2.0"
__date__ = "2022-01-26"
__updated__ = "2023-02-13"

_logger = logging.getLogger(__name__)

_PYLINT_DEST = "pylint"
"""Name of destination variable used for pylint in parsing"""


def __parser() -> argparse.ArgumentParser:
    program_version_message = f"%(prog)s {__version__} ({__updated__})"
    program_short_description = __doc__.split(".", maxsplit=1)[0]
    program_license = f"""{program_short_description}

      Created by Vincent Hatakeyama on {__date__}.
      Copyright 2022 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """
    parser = basic_parser(program_license, program_version_message)
    return parser


def parser_add_pylint_group(parser: argparse.ArgumentParser):
    """Add pylint option to the given parser.
    Defaults to not running pylint."""
    parser.add_argument(
        "--pylint",
        help="Run pylint on given python module (for example odoo.addons.base)",
        dest=_PYLINT_DEST,
        default=[],
        nargs="+",
    )
    parser.add_argument(
        "--pylint-output-format",
        help="Run pylint with the --output-format provided (for example colorized or "
        "parseable)",
        dest="pylint_output_format",
        default="colorized",
    )


def apply_pylint(namespace: argparse.Namespace, config: Configuration) -> int:
    """Run pylint if the option was set."""
    build_options = get_build_options(namespace)
    display_print_version: bool = False
    if display_print_version:
        pylint_print_version(config, build_options)
    if _PYLINT_DEST in namespace and getattr(namespace, _PYLINT_DEST):
        return pylint(
            config,
            getattr(namespace, _PYLINT_DEST),
            build_options,
            namespace.pylint_output_format,
        )
    return 0


def main(argv: Optional[List[str]] = None) -> int:
    """Copy modules for a build, callable version that parses arguments"""
    parser = __parser()
    parser_add_pylint_group(parser)
    add_build_options(parser)
    nmspc = parser.parse_args(argv)
    apply(nmspc)
    return apply_pylint(nmspc, Config())


def pylint_print_version(config: Configuration, build_options: List[str]):
    """Output pylint version"""
    build_local_image(build_options)

    # TODO change to have a single Docker run with both version and eventual pylint run
    DockerClient.run(
        config.local_image,
        config.local_tag,
        {
            "entrypoint": "pylint",
            "command": ["--version"],
        },
        False,
    )


def pylint(
    config: Configuration,
    modules: List[str],
    build_options: List[str],
    output_format: str = "colorized",
):
    """Run pylint"""

    mounts: List[Mount]
    target_source_dict: Dict[str, str]
    mounts, target_source_dict = modules_mount(config)
    environment = {}

    # Add a volume to store cache
    cache_volume_name = "pylint-cache"
    DockerClient.create_volume(cache_volume_name)
    cache_target_path = "/var/cache/pylint"
    mounts.append(mount(cache_target_path, cache_volume_name))
    environment["PYLINTHOME"] = cache_target_path

    mounts.extend(anthem_mounts(config))

    build_local_image(build_options)

    _logger.info("Running pylint on %s", modules)
    return DockerClient.run(
        config.local_image,
        config.local_tag,
        {
            "entrypoint": "pylint",
            "command": [f"--output-format={output_format}"] + modules,
            # needed to write in the cache
            "user": "root",
            "mounts": mounts,
            "envs": environment,
        },
        False,
        target_source_dict,
    )


if __name__ == "__main__":
    sys.exit(main())
