#!/usr/bin/env python3
import sys
from contextlib import closing
from getpass import getpass
from typing import Optional

# Only available if docker section is installed
try:
    import secretstorage

    __secretstorage_available = True
except ModuleNotFoundError:
    __secretstorage_available = False


__version__ = "1.0.0"
__date__ = "2023-10-23"
__updated__ = "2023-10-23"

__orus_api_token_attributes = {
    "application": "odoo_scripts",
    "server": "orus.io",
    "scheme": "https",
    "type": "api_token",
}


def get_orus_api_token() -> Optional[str]:
    """Retrieve Orus API token from secret service.

    :return: None if there is no token"""
    if not __secretstorage_available:
        return None
    with closing(secretstorage.dbus_init()) as conn:
        collection = secretstorage.get_default_collection(conn)
        for item in collection.search_items(__orus_api_token_attributes):
            return item.get_secret().decode("UTF-8")
    return None


def set_orus_api_token(token: str):
    """Store Orus API token from secret service"""
    with closing(secretstorage.dbus_init()) as conn:
        collection = secretstorage.get_default_collection(conn)
        collection.create_item(
            "Orus API Token for Odoo Scripts",
            __orus_api_token_attributes,
            token.encode("UTF-8"),
            True,
        )


def main() -> int:
    """Ask for token and set it"""
    token = getpass("Type in token value followed by enter:")
    set_orus_api_token(token)
    return 0


if __name__ == "__main__":
    sys.exit(main())
